import typing
import warnings

import pyo

import walkman


class IllegalMidiControlIndexWarning(Warning):
    pass


class Parameter(object):
    # Constants
    # 0 is value, 1 is scale object,
    # everything above are midi control inputs.
    VALUE_VOICE: int = 0
    SCALE_VOICE: int = 1

    def __init__(
        self,
        input_provider: walkman.InputProvider,
    ):
        # If value is only a numerical value, we will only
        # pass it to 'pyo.Sig'
        self._sig = pyo.Sig(0)
        # Scale is used if a midi control input is used.
        # It get initialised with dummy values
        self._scale = pyo.Scale(self._sig)
        self._input_provider = input_provider
        self._value = pyo.Selector(
            [self._sig, self._scale] + self.input_provider.midi_input_list
        )

    @property
    def input_provider(self) -> walkman.InputProvider:
        return self._input_provider

    def initialise(
        self,
        value: float = 0,
        midi_control_index: typing.Optional[int] = None,
        midi_range: typing.Optional[typing.Tuple[float, float]] = None,
    ):
        selector_voice = None
        if midi_control_index is not None:
            try:
                value_control = self.input_provider.midi_input_list[midi_control_index]
            except IndexError:
                warnings.warn(
                    f"Found invalid midi_control_index = {midi_control_index}."
                    f" Only values from 0 to {len(self.input_provider.midi_input_list)}"
                    " are allowed. Ignored midi data.",
                    IllegalMidiControlIndexWarning,
                )
            else:
                selector_voice = 2 + midi_control_index

            if selector_voice and midi_range is not None:
                self._scale.setInput(value_control)
                self._scale.setInMin(value_control.minscale)
                self._scale.setInMax(value_control.maxscale)
                self._scale.setOutMin(midi_range[0])
                self._scale.setOutMax(midi_range[1])
                selector_voice = self.SCALE_VOICE

        if selector_voice is None:
            self._sig.setValue(value)
            value_control = self._sig
            selector_voice = self.VALUE_VOICE

        self._value.setVoice(selector_voice)

    @property
    def value(self) -> pyo.PyoObject:
        return self._value



import functools
import typing
import warnings

import walkman

ModuleNameToInitialiseKwargsDict = typing.Dict[str, typing.Union[bool, dict]]


class Cue(object):
    def __init__(
        self,
        module_dict: walkman.ModuleDict,
        name: str,
        **kwargs: typing.Dict[int, typing.Union[dict, bool]],
    ):
        self._is_playing = False
        self._module_dict = module_dict
        self._name = name
        self._module_name_to_replication_configuration = kwargs
        # We can't call '_set_duration' when initialising a Cue,
        # because the cue won't be activated yet and the modules
        # will return wrong values for 'duration'.
        self._duration = None

    def _set_duration(self):
        if self._duration is None:
            duration_list = []
            for module_name in self.active_module_name_set:
                for module in self._module_dict[module_name]:
                    duration_list.append(module.duration)
            try:
                duration = max(duration_list)
            except ValueError:
                duration = 0
            self._duration = duration

    def _initialise_module_tuple(
        self,
        module_tuple: typing.Tuple[walkman.Module, ...],
        replication_configuration: typing.Dict[int, typing.Dict[str, typing.Any]],
    ):
        for (
            module_replication_index,
            initialise_kwargs,
        ) in replication_configuration.items():
            try:
                module = module_tuple[module_replication_index]
            except IndexError:
                warnings.warn(
                    f"Module {module_name} only has {len(module_tuple)} "
                    f"replications. Replication '{module_replication_index}'"
                    f" doesn't exist."
                )
            else:
                try:
                    module.initialise(**initialise_kwargs)
                # Instead of providing kwargs, we can also simply
                # write 'false', to stop modules with 'auto_stop = false'
                except TypeError:
                    if initialise_kwargs is False:
                        module.stop()

    @functools.cached_property
    def active_module_name_set(self) -> typing.Set[str]:
        return set(
            module_name
            for module_name in self.module_name_to_replication_configuration.keys()
            if module_name in self._module_dict
        )

    @functools.cached_property
    def active_module_tuple(self) -> typing.Tuple[walkman.Module, ...]:
        active_module_list = []
        for module_name in self.active_module_name_set:
            for module in self._module_dict[module_name]:
                active_module_list.append(module)
        return tuple(active_module_list)

    @functools.cached_property
    def name(self) -> str:
        return self._name

    @property
    def duration(self) -> float:
        if self._duration:
            return self._duration
        return 0

    @property
    def is_playing(self) -> bool:
        return self._is_playing

    @property
    def module_name_to_replication_configuration(
        self,
    ) -> ModuleNameToInitialiseKwargsDict:
        return self._module_name_to_replication_configuration

    def play(self):
        for module in self.active_module_tuple:
            module.play()
        self._is_playing = True

    def stop(self):
        for module in self.active_module_tuple:
            module.stop()
        self._is_playing = False

    def activate(self):
        for module_name, module_tuple in self._module_dict.items():
            try:
                replication_configuration = (
                    self.module_name_to_replication_configuration[module_name]
                )
            except KeyError:
                for module in module_tuple:
                    if module.auto_stop:
                        module.stop()
            else:
                self._initialise_module_tuple(module_tuple, replication_configuration)

            self._set_duration()

    def jump_to(self, time_in_seconds: float):
        for module in self.active_module_tuple:
            module.jump_to(time_in_seconds)


class CueManager(object):
    def __init__(self, cue_tuple: typing.Tuple[Cue, ...]):
        self._cue_index = 0
        self._cue_tuple = cue_tuple
        self._cue_count = len(cue_tuple)

        self._active_cue_index = None
        self._activate_cue()

        self._cue_name_tuple = tuple(cue.name for cue in self._cue_tuple)

    def _set_cue_index(self, index: int):
        self._cue_index = index % self._cue_count

    def _activate_cue(self):
        if self._active_cue_index != self.current_cue_index:
            self.current_cue.activate()
            self._active_cue_index = self.current_cue_index
        self.current_cue.jump_to(0)

    def _move_and_activate_cue(self, cue_index: int):
        if is_playing := self.current_cue.is_playing:
            self.current_cue.stop()
        self._set_cue_index(cue_index)
        self._activate_cue()
        if is_playing:
            self.current_cue.play()

    def __getitem__(self, index_or_key: typing.Union[str, int]) -> Cue:
        if isinstance(index_or_key, str):
            index = self._cue_name_tuple.index(index_or_key)
        else:
            index = index_or_key
        return self._cue_tuple[index]

    @property
    def cue_name_tuple(self) -> typing.Tuple[str, ...]:
        return self._cue_name_tuple

    @property
    def active_module_tuple(self):
        return self.current_cue.active_module_tuple

    @property
    def cue_count(self):
        return self._cue_count

    @property
    def current_cue_index(self):
        return self._cue_index

    @property
    def current_cue(self) -> Cue:
        return self._cue_tuple[self._cue_index]

    def move_to_previous_cue(self):
        self._move_and_activate_cue(self._cue_index - 1)

    def move_to_next_cue(self):
        self._move_and_activate_cue(self._cue_index + 1)

    def jump_to_cue(self, cue_index: int):
        self._move_and_activate_cue(cue_index)



from .version import __version__
from . import constants
from . import utilities
from .io import (
    InputProvider,
    OutputProvider,
    ChannelMapping,
    dict_or_channel_mapping_to_channel_mapping,
)
from .audio import SimpleAudioObject, NestedAudioObject, AudioHost
from .parameters import Parameter
from .modules import (
    Module,
    ModuleWithDecibel,
    ModuleWithDecibelControlledAutoStartStop,
    ModuleDict,
)
from .cues import Cue, CueManager
from . import tests
from .backends import Backend
from . import parsers
from . import ui
from .boot import start_loop_from_jinja2_file_path, start_loop_from_toml_str



from __future__ import annotations

import abc

import pyo

import walkman


class AudioObject(abc.ABC):
    @classmethod
    def get_name(cls) -> str:
        return walkman.utilities.camel_case_to_snake_case(cls.__name__)

    def close(self):
        """Method is called when walkman is closed"""
        pass


class SimpleAudioObject(AudioObject):
    @abc.abstractproperty
    def pyo_object(self) -> pyo.PyoObject:
        ...


class NestedAudioObject(AudioObject):
    @abc.abstractproperty
    def pyo_object_list(self) -> list[pyo.PyoObject]:
        ...


class AudioHost(object):
    """Wrapper for pyo.Server and pyo.Mixer.

    Simplifies server API and adds volumes controls.
    """

    def __init__(
        self,
        audio: str = "jack",
        midi: str = "jack",
        sampling_rate: int = 44100,
        buffer_size: int = 256,
        output_channel_mapping: walkman.ChannelMapping = walkman.ChannelMapping(
            {0: 0, 1: 1}
        ),
    ):
        self._output_channel_mapping = output_channel_mapping
        self._is_playing = False
        self.server = pyo.Server(
            sr=sampling_rate,
            midi=midi,
            nchnls=output_channel_mapping.maxima_right_channel,
            buffersize=buffer_size,
            duplex=1,
            audio=audio,
            jackname=walkman.constants.NAME,
        ).boot()

    @property
    def output_channel_mapping(self) -> walkman.ChannelMapping:
        return self._output_channel_mapping

    @property
    def is_playing(self) -> bool:
        return self._is_playing

    def start(self):
        self.server.start()
        self._is_playing = True

    def stop(self):
        self.server.stop()
        self._is_playing = False

    def close(self):
        self.stop()
        del self.server



"""This module handles input/output data of walkman."""

from __future__ import annotations
import dataclasses
import typing
import warnings

import pyo

import walkman


class ChannelMapping(typing.Dict[int, typing.Tuple[int, ...]]):
    """Map audio channels to other audio channels.

    Left side is always one channel, right side can be one
    or more channels.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Hack to ensure key/values are integer
        key_to_remove_list = []
        update_dict = {}
        for key, value in self.items():
            if not isinstance(key, int):
                key_to_remove_list.append(key)
            if isinstance(value, (str, int)):
                output_channel_list = [value]
            else:
                output_channel_list = value
            output_channel_tuple = tuple(
                int(output_channel) for output_channel in output_channel_list
            )
            update_dict[int(key)] = output_channel_tuple
        self.update(update_dict)
        for key in key_to_remove_list:
            del self[key]

    @property
    def left_channel_set(self) -> typing.Set[int]:
        return set(self.keys())

    @property
    def maxima_left_channel(self) -> int:
        return max(self.left_channel_set) + 1

    @property
    def right_channel_tuple_tuple(self) -> typing.Tuple[typing.Tuple[int, ...], ...]:
        return tuple(self.values())

    @property
    def right_channel_set(self) -> typing.Set[int]:
        right_channel_set = set([])
        for right_channel_tuple in self.right_channel_tuple_tuple:
            for right_channel in right_channel_tuple:
                right_channel_set.add(right_channel)
        return right_channel_set

    @property
    def maxima_right_channel(self) -> int:
        return max(self.right_channel_set) + 1

    def to_mixer(self) -> pyo.Mixer:
        return pyo.Mixer(outs=self.maxima_right_channel)


def dict_or_channel_mapping_to_channel_mapping(
    dict_or_channel_mapping_to_channel_mapping: typing.Union[dict, ChannelMapping]
) -> ChannelMapping:
    if not isinstance(dict_or_channel_mapping_to_channel_mapping, ChannelMapping):
        channel_mapping = ChannelMapping(dict_or_channel_mapping_to_channel_mapping)
    else:
        channel_mapping = dict_or_channel_mapping_to_channel_mapping
    return channel_mapping


MidiControlNumber = int
MidiChannel = int
MidiControlList = typing.List[typing.Tuple[MidiControlNumber, MidiChannel]]
MidiInputList = typing.List[pyo.Midictl]

AudioInputList = typing.List[pyo.PyoObject]


@dataclasses.dataclass(frozen=True)
class InputProvider(object):
    """Collect all physical audio and control inputs"""

    audio_input_list: AudioInputList
    midi_input_list: MidiInputList

    @staticmethod
    def channel_mapping_to_audio_input_list(
        channel_mapping: ChannelMapping,
    ) -> AudioInputList:
        if not channel_mapping:
            return []

        physical_channel_index_to_input_dict = {}
        for physical_channel in channel_mapping.keys():
            pyo_input = pyo.Input(chnl=physical_channel)
            pyo_input.play()
            physical_channel_index_to_input_dict.update(
                {physical_channel: pyo_input}
            )

        denormal_noise = pyo.Noise(1e-24)
        denormal_noise.play()

        audio_input_list = [
            denormal_noise for _ in range(channel_mapping.maxima_left_channel)
        ]

        for physical_channel, audio_input_index_tuple in channel_mapping.items():
            pyo_input = physical_channel_index_to_input_dict[physical_channel]
            for audio_input_index in audio_input_index_tuple:
                audio_input_list[audio_input_index] += pyo_input

        return audio_input_list

    @staticmethod
    def midi_control_list_to_midi_input_list(
        midi_control_list: MidiControlList,
    ) -> MidiInputList:
        midi_input_list = []
        for midi_control_number, midi_channel in midi_control_list:
            midi_ctl = pyo.Midictl(midi_control_number, channel=midi_channel)
            midi_ctl.play()
            midi_input_list.append(midi_ctl)
        return midi_input_list

    @classmethod
    def from_data(
        cls,
        channel_mapping: typing.Union[dict, ChannelMapping] = ChannelMapping({}),
        midi_control_list: MidiControlList = [],
    ) -> InputProvider:
        channel_mapping = dict_or_channel_mapping_to_channel_mapping(channel_mapping)
        audio_input_list = cls.channel_mapping_to_audio_input_list(channel_mapping)
        midi_input_list = cls.midi_control_list_to_midi_input_list(midi_control_list)
        return cls(audio_input_list=audio_input_list, midi_input_list=midi_input_list)


class IllegalChannelIndexWarning(Warning):
    """Call if user gave too high channel"""


MixerIndex = int
MixerInfo = typing.Tuple[MixerIndex, ...]
"""Each channel of an audio signal has a specific index in the mixer"""


class PyoObjectMixer(pyo.PyoObject):
    def __init__(self, base_objs: list):
        super().__init__()
        self._base_objs = base_objs


@dataclasses.dataclass
class OutputProvider(object):
    """Collect all outputs from modules and send them to physical outputs"""

    channel_mapping: ChannelMapping = dataclasses.field(
        default_factory=lambda: ChannelMapping({0: 0, 1: 1})
    )
    output_mixer: pyo.Mixer = dataclasses.field(init=False)
    module_mixer: pyo.Mixer = dataclasses.field(init=False)
    module_channel_count: int = dataclasses.field(init=False)
    physical_output_channel_count: int = dataclasses.field(init=False)
    name_to_mixer_info: typing.Dict[str, MixerInfo] = dataclasses.field(
        default_factory=lambda: {}
    )

    def __post_init__(self):
        self.channel_mapping = dict_or_channel_mapping_to_channel_mapping(
            self.channel_mapping
        )
        self.module_channel_count = self.channel_mapping.maxima_left_channel
        self.physical_output_channel_count = self.channel_mapping.maxima_right_channel
        self.module_mixer = pyo.Mixer(outs=self.module_channel_count)
        self.output_mixer = pyo.Mixer(outs=self.physical_output_channel_count)

        for (
            module_channel_index,
            physical_output_channel_index_tuple,
        ) in self.channel_mapping.items():
            pyo_object = PyoObjectMixer(self.module_mixer[module_channel_index])
            self.output_mixer.addInput(module_channel_index, pyo_object)
            for physical_output_channel_index in physical_output_channel_index_tuple:
                self.output_mixer.setAmp(
                    module_channel_index, physical_output_channel_index, 1
                )

        for channel_index, channel in enumerate(self.output_mixer):
            channel[0].out(chnl=channel_index)

    def _raise_illegal_audio_object_channel_warning(
        self,
        audio_object: walkman.AudioObject,
        channel_mapping: ChannelMapping,
        audio_object_channel_index: int,
    ):
        warnings.warn(
            "Found invalid audio_object_channel_index ="
            f" {audio_object_channel_index} in channel_mapping ="
            f" {channel_mapping} for audio object '{audio_object.name}'."
            f" Audio object only have {len(audio_object.pyo_object)}"
            " channels.",
            IllegalChannelIndexWarning,
        )

    def register_audio_object(self, simple_audio_object: walkman.SimpleAudioObject) -> MixerInfo:
        mixer_info = []
        base_index = walkman.utilities.get_next_mixer_index(self.module_mixer)
        for audio_object_index in range(len(simple_audio_object.pyo_object)):
            pyo_channel = simple_audio_object.pyo_object[audio_object_index]
            mixer_index = base_index + audio_object_index
            self.module_mixer.addInput(mixer_index, pyo_channel)
            mixer_info.append(mixer_index)
        return tuple(mixer_info)

    def activate_channel_mapping(
        self, simple_audio_object: walkman.SimpleAudioObject, channel_mapping: ChannelMapping
    ):
        name = simple_audio_object.get_name()
        try:
            mixer_info = self.name_to_mixer_info[name]
        except KeyError:
            mixer_info = self.register_audio_object(simple_audio_object)

        for (
            audio_object_channel_index,
            output_channel_index_tuple,
        ) in channel_mapping.items():
            try:
                mixer_index = mixer_info[audio_object_channel_index]
            except IndexError:
                self._raise_illegal_audio_object_channel_warning(
                    simple_audio_object, channel_mapping, audio_object_channel_index
                )
            else:
                for module_output_channel_index in range(
                    self.module_channel_count
                ):
                    amplitude = int(module_output_channel_index in output_channel_index_tuple)
                    self.module_mixer.setAmp(
                        mixer_index, module_output_channel_index, amplitude
                    )



import dataclasses
import typing

import walkman


@dataclasses.dataclass(frozen=True)
class Backend(object):
    """Namespace for all relevant backend objects"""

    name: str
    audio_host: walkman.AudioHost
    input_provider: walkman.InputProvider
    output_provider: walkman.OutputProvider
    module_dict: walkman.ModuleDict
    cue_manager: walkman.CueManager

    def get_audio_test(
        self,
        audio_test_class: typing.Type[
            walkman.tests.AudioTest
        ] = walkman.tests.AudioRotationTest,
    ) -> walkman.tests.AudioTest:
        return audio_test_class(self.audio_host.output_channel_mapping)



import pyo


def camel_case_to_snake_case(camel_case_string: str) -> str:
    """Transform camel case formatted string to snake case.

    :param camel_case_string: String which is formatted using
        camel case (no whitespace, but upper letters at
        new word start).
    :return: string formatted using snake case

    **Example:** MyClassName -> my_class_name
    """

    character_list = []

    is_first = True
    for character in camel_case_string:
        if character.isupper():
            character = character.lower()
            if not is_first:
                character = "_{}".format(character)

        if is_first:
            is_first = False

        character_list.append(character)

    return "".join(character_list)


def decibel_to_amplitude_ratio(decibel: float, reference_amplitude: float = 1) -> float:
    """Convert decibel to amplitude ratio.

    :param decibel: The decibel number that shall be converted.
    :param reference_amplitude: The amplitude for decibel == 0.

    """
    return float(reference_amplitude * (10 ** (decibel / 20)))


def get_next_mixer_index(mixer: pyo.Mixer) -> int:
    try:
        index = max(mixer.getKeys()) + 1
    except ValueError:
        index = 0
    return index



import setuptools  # type: ignore

version = {}
with open("walkman/version.py") as fp:
    exec(fp.read(), version)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="audiowalkman",
    version=version["__version__"],
    license="GPL",
    description="play audio files in performance contexts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Levin Eric Zimmermann",
    author_email="levin.eric.zimmermann@posteo.eu",
    packages=["walkman"],
    setup_requires=[],
    install_requires=[
        # for audio
        "pyo==1.0.4",
        # for GUI
        "PySimpleGUI>=4.60.0, <5.0.0",
        # for CLI
        "click>=8.1.3, <9.0.0",
        # to read config files
        "tomli>=2.0.1, <3.0.0",
        # to parse jinja2 syntax in config files
        "jinja2>=3.1.2, <4.0.0",
    ],
    extras_require={"testing": ["nose"]},
    python_requires="==3.8",
    entry_points={"console_scripts": ["walkman=walkman.__main__:main"]},
)



"""Define interface for modules"""

from __future__ import annotations

import abc
import copy
import dataclasses
import functools
import importlib
import inspect
import pkgutil
import typing

import pyo

import walkman


def override_default_kwargs(method_to_wrap: typing.Callable) -> typing.Callable:
    def wrapper(self, **kwargs):
        default_dict = copy.deepcopy(self.default_dict)
        default_dict.update(kwargs)
        return method_to_wrap(self, **default_dict)

    return wrapper


@dataclasses.dataclass
class Module(walkman.SimpleAudioObject):
    """Interface for an isolated audio process."""

    audio_host: walkman.AudioHost
    input_provider: walkman.InputProvider
    output_provider: walkman.OutputProvider
    auto_stop: bool = True
    fade_in_duration: float = 0.1
    fade_out_duration: float = 0.2
    default_dict: typing.Dict[str, typing.Any] = dataclasses.field(
        default_factory=lambda: {}
    )
    parameter_name_to_parameter_dict: typing.Dict[
        str, walkman.Parameter
    ] = dataclasses.field(default_factory=lambda: {})
    internal_pyo_object_list: typing.List[pyo.PyoObject] = dataclasses.field(
        default_factory=lambda: []
    )

    def __post_init__(self):
        self._is_playing = False
        self.fader = pyo.Fader(self.fade_in_duration, self.fade_out_duration)
        self.setup_pyo_object()
        self._pyo_object_with_fader = self._pyo_object * self.fader

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.get_name()})"

    @staticmethod
    def value_or_parameter_kwargs_to_parameter_kwargs(
        value_or_parameter_kwargs: typing.Union[dict, float]
    ) -> dict:
        if isinstance(value_or_parameter_kwargs, dict):
            parameter_kwargs = value_or_parameter_kwargs
        else:
            parameter_kwargs = {"value": value_or_parameter_kwargs}
        return parameter_kwargs

    @functools.cached_property
    def default_parameter_name_to_parameter_kwargs_dict(
        self,
    ) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
        parameter_name_to_parameter_kwargs_dict = {}
        parameter_mapping_proxy = inspect.signature(self._initialise).parameters
        for parameter_name, inspect_parameter in parameter_mapping_proxy.items():
            if inspect_parameter.annotation == "walkman.Parameter":
                parameter_name_to_parameter_kwargs_dict.update(
                    {
                        parameter_name: self.value_or_parameter_kwargs_to_parameter_kwargs(
                            inspect_parameter.default
                        )
                    }
                )

        return parameter_name_to_parameter_kwargs_dict

    def _fetch_keyword_argument_dict(self, **kwargs) -> typing.Dict[str, typing.Any]:
        keyword_argument_dict = {}

        parameter_name_to_parameter_kwargs_dict = copy.deepcopy(
            self.default_parameter_name_to_parameter_kwargs_dict
        )

        for argument, value in kwargs.items():
            if argument in parameter_name_to_parameter_kwargs_dict:
                parameter_name_to_parameter_kwargs_dict[
                    argument
                ] = self.value_or_parameter_kwargs_to_parameter_kwargs(value)
            else:
                keyword_argument_dict.update({argument: value})

        for (
            parameter_name,
            parameter_kwargs,
        ) in parameter_name_to_parameter_kwargs_dict.items():
            try:
                parameter = self.parameter_name_to_parameter_dict[parameter_name]
            except KeyError:
                parameter = walkman.Parameter(self.input_provider)
                self.parameter_name_to_parameter_dict.update(
                    {parameter_name: parameter}
                )
            parameter.initialise(**parameter_kwargs)
            keyword_argument_dict.update({parameter_name: parameter})

        return keyword_argument_dict

    # ################## ABSTRACT PROPERTIES ################## #

    @abc.abstractproperty
    def _pyo_object(self) -> pyo.PyoObject:
        ...

    # ################## ABSTRACT METHODS ################## #

    @abc.abstractmethod
    def setup_pyo_object(self):
        ...

    @abc.abstractmethod
    def _initialise(self, **kwargs):  # type: ignore
        ...

    def _play(self, duration: float = 0, delay: float = 0):
        for internal_pyo_object in self.internal_pyo_object_list:
            internal_pyo_object.play(dur=duration, delay=delay)

    def _stop(self, wait: float = 0):
        for internal_pyo_object in self.internal_pyo_object_list:
            internal_pyo_object.stop(wait=wait)

    # ################## PUBLIC PROPERTIES ################## #

    @property
    def pyo_object(self) -> pyo.PyoObject:
        return self._pyo_object_with_fader

    @property
    def duration(self) -> float:
        return 0

    @property
    def is_playing(self) -> bool:
        return self._is_playing

    # ################## PUBLIC METHODS ################## #

    @override_default_kwargs
    def initialise(
        self,
        channel_mapping: typing.Optional[
            typing.Union[dict, walkman.ChannelMapping]
        ] = None,
        **kwargs,
    ):
        # Activate the channel mapping
        if channel_mapping is None:
            channel_mapping = walkman.ChannelMapping(
                {channel: channel for channel in range(len(self.pyo_object))}
            )
        else:
            channel_mapping = walkman.dict_or_channel_mapping_to_channel_mapping(
                channel_mapping
            )
        self.output_provider.activate_channel_mapping(self, channel_mapping)

        # Fetch parameters
        keyword_argument_dict = self._fetch_keyword_argument_dict(**kwargs)

        # Parse everything to actual initialise method
        self._initialise(**keyword_argument_dict)

    def play(self, duration: float = 0, delay: float = 0) -> Module:
        if not self.is_playing:
            self.fader.play(dur=duration, delay=delay)
            self._play(duration=duration, delay=delay)
            self._is_playing = True
        return self

    def stop(self, wait: float = 0) -> Module:
        if self.is_playing:
            self.fader.stop(wait)
            self._stop(wait + self.fade_out_duration)
            self._is_playing = False
        return self

    def close(self):
        self.stop()

    def jump_to(self, time_in_seconds: float):
        """To jump at certain points of the sound process.

        Can optionally be implemented.
        """

        pass


class ModuleWithDecibel(Module):
    def setup_pyo_object(self):
        # Default audio objects, used to control default parameter
        # 'decibel'
        self._decibel_value = pyo.Sig(0)
        self._decibel_to_amplitude = pyo.DBToA(self._decibel_value)
        self._decibel_signal_to = pyo.SigTo(self._decibel_to_amplitude, time=0.015)

        # XXX: We SHOULD NOT add self._decibel_value to 'internal_pyo_object_list', because
        # 'ModuleWithDecibelControlledAutoStartStop' needs a running function to detect
        # of module should be stopped.
        self.internal_pyo_object_list.extend(
            [self._decibel_to_amplitude, self._decibel_signal_to]
        )

    def _initialise(self, decibel: walkman.Parameter = -6, **kwargs):  # type: ignore
        self._decibel_value.setValue(decibel.value)

    def play(self, duration: float = 0, delay: float = 0) -> Module:
        self._decibel_value.play(dur=duration, delay=delay)
        super().play(duration, delay)
        return self

    def stop(self, wait: float = 0) -> Module:
        super().stop(wait)
        self._decibel_value.stop(wait + self.fade_out_duration)
        return self


@dataclasses.dataclass()
class ModuleWithDecibelControlledAutoStartStop(ModuleWithDecibel):
    decibel_threshold: float = -100

    def setup_pyo_object(self):
        super().setup_pyo_object()

        # Initialise auto-start / auto-stop triggers, to improve performance
        # of program (if module is so quiet that it can't be heard it
        # should be stopped).
        self._rising_decibel_tracker = pyo.Thresh(
            self._decibel_value, self.decibel_threshold, dir=0
        ).play()
        self._auto_start_trigger = pyo.TrigFunc(
            self._rising_decibel_tracker, self._play
        ).play()

        self._falling_decibel_tracker = pyo.Thresh(
            self._decibel_value, self.decibel_threshold, dir=1
        ).play()
        self._auto_stop_trigger = pyo.TrigFunc(
            self._falling_decibel_tracker, self._stop
        ).play()

        # XXX: We don't add the trigger and tracker objects
        # to the `internal_pyo_object_list`, because they should always play
        # and never be stopped!
        self._auto_start_stop_pyo_object_list = [
            self._rising_decibel_tracker,
            self._auto_start_trigger,
            self._falling_decibel_tracker,
            self._auto_stop_trigger,
        ]

    def play(self, duration: float = 0, delay: float = 0) -> Module:
        super().play(duration, delay)
        for auto_start_stop_pyo_object in self._auto_start_stop_pyo_object_list:
            auto_start_stop_pyo_object.play(dur=duration, delay=delay)
        return self

    def stop(self, wait: float = 0) -> Module:
        super().stop(wait)
        for auto_start_stop_pyo_object in self._auto_start_stop_pyo_object_list:
            auto_start_stop_pyo_object.stop(wait)
        return self


ModuleName = str


class ModuleDict(typing.Dict[ModuleName, typing.Tuple[Module, ...]]):
    @staticmethod
    def get_module_name_to_module_class_dict():
        namespace_package = __import__(walkman.constants.MODULE_PACKAGE_NAME)
        module_info_list = list(
            pkgutil.iter_modules(
                namespace_package.__path__, namespace_package.__name__ + "."
            )
        )
        module_name_to_module_class_dict = {}
        for module_info in module_info_list:
            module = importlib.import_module(module_info.name)
            for _, python_class in inspect.getmembers(module, inspect.isclass):
                if Module in inspect.getmro(python_class):
                    module_name_to_module_class_dict.update(
                        {python_class.get_name(): python_class}
                    )
        return module_name_to_module_class_dict

    @classmethod
    def from_audio_objects_and_module_configuration(
        cls,
        audio_host: walkman.AudioHost,
        input_provider: walkman.InputProvider,
        output_provider: walkman.OutputProvider,
        module_name_to_replication_configuration_dict: typing.Dict[
            str, typing.Dict[int, typing.Dict[str, typing.Any]]
        ],
    ) -> ModuleDict:
        module_name_to_module_class_dict = cls.get_module_name_to_module_class_dict()

        module_name_to_module_dict = {}
        for module_name, module_class in module_name_to_module_class_dict.items():
            try:
                replication_index_to_module_configuration_dict = (
                    module_name_to_replication_configuration_dict[module_name]
                )
            except KeyError:
                replication_index_to_module_configuration_dict = {0: {}}
            module_list = [
                None
                for _ in range(
                    max(replication_index_to_module_configuration_dict.keys()) + 1
                )
            ]
            for (
                replication_index,
                module_configuration,
            ) in replication_index_to_module_configuration_dict.items():
                module = module_class(
                    audio_host, input_provider, output_provider, **module_configuration
                )
                module_list[replication_index] = module
            module_tuple = tuple(
                module
                if module is not None
                else module_class(audio_host, input_provider, output_provider)
                for module in module_list
            )
            module_name_to_module_dict.update({module_name: module_tuple})
        return cls(module_name_to_module_dict)

    def close(self):
        for module in self.module_tuple:
            module.close()

    @functools.cached_property
    def module_tuple(self) -> typing.Tuple[walkman.Module, ...]:
        module_list = []
        for module_tuple in self.values():
            module_list.extend(module_tuple)
        return tuple(module_list)



import click

import walkman

@click.command()
@click.argument("configuration_file_path")
def main(configuration_file_path: str):
    walkman.start_loop_from_jinja2_file_path(configuration_file_path)



"""Boot walkman loop (with or without UI)"""

import walkman


def start_loop(
    backend: walkman.Backend,
    headless: bool = False,
):
    if not headless:
        gui = walkman.ui.backend_to_gui(backend)
        gui.loop()
    else:
        while True:
            pass


def start_loop_from_jinja2_file_path(jinja2_file_path: str, **kwargs):
    backend = walkman.parsers.jinja2_file_path_to_backend(jinja2_file_path)
    start_loop(backend, **kwargs)


def start_loop_from_toml_str(toml_file_path: str, **kwargs):
    backend = walkman.parsers.toml_str_to_backend(toml_file_path)
    start_loop(backend, **kwargs)


